import logging, re, threading, os, errno
import select
import shutil
import plesk_config, gzip_stream

import tempfile, tempfile_compat


try:
    import subprocess
except ImportError:
    import subprocess_compat
    subprocess = subprocess_compat

_logger = logging.getLogger("api.unpacker")

class Id(object):
    def __init__(self, str_id):
        res = re.search("(.+):gzipped$", str_id)
        if res:
            str_id = res.group(1)
            self.gzipped = True
        else:
            self.gzipped = False

        self.filename = str_id


class Unpacker(object):
    def unpack(self, fd, str_id):
        id = Id(str_id)

        if id.gzipped:
            fd = gzip_stream.GzipReadStream(fd)

        return self.unpack_stream(fd)


def _write_mysql_config_file(fd, user, password):
    template = \
"""[client]
user=%(user)s
password=%(password)s
"""
    os.write(fd, template % {'user': user, 'password': password})
    os.close(fd)

def pump(cmd, fd, env = None):
    sub = subprocess.Popen(cmd, stdin = subprocess.PIPE,
                           stdout = subprocess.PIPE, stderr = subprocess.PIPE, env = env)

    read_set = [sub.stdout, sub.stderr]
    write_set = [sub.stdin]

    stdout = ""
    stderr = ""

    stdin_buffer = ""

    while read_set or write_set:
        #_logger.debug(str([x.fileno() for x in read_set]) + " $ " + str([x.fileno() for x in write_set]) + " ->")
        rlist, wlist, xlist = select.select(read_set, write_set, [])
        #_logger.debug(str([x.fileno() for x in rlist]) + " $ " + str([x.fileno() for x in wlist]))

        # 512-bytes-writes are guaranteed to not block
        # if select returned 'available to write' status
        # on descriptor, so all writes go in the 512-byte chunks.
        #
        # 4096 is the optimal size of fd reading due to internals of MIME-extractor

        if sub.stdout in rlist:
            data = os.read(sub.stdout.fileno(), 1024)
            if data == "":
                sub.stdout.close()
                read_set.remove(sub.stdout)
            #    _logger.debug("1X")
            #else:
            #    _logger.debug("1>" + data)
            stdout += data
            continue

        if sub.stderr in rlist:
            data = os.read(sub.stderr.fileno(), 1024)
            if data == "":
                sub.stderr.close()
                read_set.remove(sub.stderr)
            #    _logger.debug("2X")
            #else:
            #    _logger.debug("2>" + data)
            stderr += data
            continue

        if sub.stdin in wlist:
            if len(stdin_buffer) < 512:
                if fd:
                    data = fd.read(4096)
                    if data == "":
                        fd = None
                    stdin_buffer += data

            if not len(stdin_buffer):
                sub.stdin.close()
                write_set.remove(sub.stdin)
            else:
                try:
                    written = os.write(sub.stdin.fileno(), stdin_buffer[:512])
                    #_logger.debug("0<" + str(len(written)))
                except OSError, e:
                    if e.errno == errno.EPIPE:
                        #_logger.debug("0<EPIPE")
                        sub.stdin.close()
                        write_set.remove(sub.stdin)
                    else:
                        raise
                else:
                    stdin_buffer = stdin_buffer[written:]


    retvalue = sub.wait()

    return (retvalue, stdout, stderr)

class DbUnpacker(Unpacker):
    def __init__(self, name, type, user, password, host, port):
        self.name = name
        self.type = type
        self.user = user
        self.password = password
        self.host = host
        self.port = port

    def unpack_stream(self, fd):
        if self.type == "mysql":
            (fh, configfile) = tempfile.mkstemp()
            try:
                _write_mysql_config_file(fh, self.user, self.password)

                cmd = [os.path.join(plesk_config.get("MYSQL_BIN_D"), "mysql")]
                cmd.append("--defaults-file=" + configfile)
                if self.host != "":
                    cmd.append("-h")
                    cmd.append(self.host)
                if self.port != "":
                    cmd.append("-P")
                    cmd.append(self.port)
                cmd.append(self.name)

                retvalue, stdout, stderr = pump(cmd, fd)

                if retvalue != 0:
                    raise ValueError(retvalue, stdout, stderr)
            finally:
                os.unlink(configfile)
        else:
            env = os.environ.copy()
            env["PGUSER"] = self.user
            env["PGPASSWORD"] = self.password

            # option --use-set-session-authorization for old version of postgresql dump, bug 90676
            cmd = [os.path.join(plesk_config.get("PGSQL_BIN_D"), "pg_restore"), "-d", self.name, "--use-set-session-authorization"]
            if self.host != "" and self.host != "localhost":
                cmd.append("-h")
                cmd.append(self.host)
                if self.port != "":
                    cmd.append("-p")
                    cmd.append(self.port)

            try:
                retvalue, stdout, stderr = pump(cmd, fd, env)
            except OSError, e:
                if e.errno == errno.ENOENT:
                    raise RuntimeError("No postgresql command-line utility installed")
                raise

            if retvalue != 0:
                raise ValueError(retvalue, stdout, stderr)

class TarUnpacker(Unpacker):
    def __init__(self, destdir):
        self.destdir = destdir

    def unpack_stream(self, fd):

        retvalue, stdout, stderr = pump([plesk_config.get("TAR_BIN"), "-xvf", "-", "-C", self.destdir], fd)

        if retvalue != 0:
            raise ValueError(retvalue, stdout, stderr)

        # TAR_BIN might be GNU or BSD tar.
        return parseGnuTarOutput(stdout) + parseBsdTarOutput(stderr)

# GNU tar: stdout, line per file
def parseGnuTarOutput(s):
    return s.splitlines()

# BSD tar: stderr, file prepended with 'x ' string
def parseBsdTarOutput(s):
    return [line[2:] for line in s.splitlines() if line.startswith('x ')]

class PartUnpacker(Unpacker):
    def __init__(self, destfile):
        self.destfile = destfile

    def unpack_stream(self, fd):
        dest = open(self.destfile, "wb")

        try:
            shutil.copyfileobj(fd, dest)
        finally:
            dest.close()


if __name__ == '__main__':
    import sys

    id = Id(sys.argv[1])
    f = file(id.filename)
    files = TarUnpacker(sys.argv[2]).unpack(f, sys.argv[1])

    for file in files:
        print file
